import sys
import inspect
import logging
import threading
from collections import namedtuple
from odoo import api, SUPERUSER_ID, _
from odoo.sql_db import BaseCursor
from odoo.http import request
from odoo.addons.saas_worker.const import ENV, TEST

_logger = logging.getLogger(__name__)

LazyRef = namedtuple('LazyRef', ['xmlid', 'default'])


def _get_env():
    try:
        return request.env
    except RuntimeError:
        # request is unbound, check a few frames back if we can found an
        # existing environment
        frame = inspect.currentframe()
        if not frame or not frame.f_back:
            return None
        frame = frame.f_back  # not need to check for env in current method
        cr = None
        for _level in range(6):  # check up to 6. frames back
            if not frame:
                break
            cr = _._get_cr(frame, allow_create=False)[0]
            if cr:
                if not isinstance(cr, BaseCursor):
                    _logger.error(type(cr), type(cr).__mro__)
                return api.Environment(cr, SUPERUSER_ID, {})
            frame = frame.f_back
        return None

class Consts(object):
    SIZE_5 = LazyRef('openerp_portal.crm_tag_size_5', 63)
    SIZE_20 = LazyRef('openerp_portal.crm_tag_size_20', 23)
    SIZE_50 = LazyRef('openerp_portal.crm_tag_size_50', 62)
    SIZE_250 = LazyRef('openerp_portal.crm_tag_size_250', 32)
    SIZE_MORE = LazyRef('openerp_portal.crm_tag_size_more', 148)

    MIN_LEADS_BY_COUNTRY = 12

    WON_STAGE = LazyRef('crm.stage_lead4', 224)

    CHANNEL_ALL_EMPLOYEES = LazyRef('mail.channel_all_employees', 1)

    #
    # Partner grades
    #
    LEARNING_GRADE = LazyRef('openerp_portal.res_partner_grade_learning', 10)
    READY_GRADE = LazyRef('openerp_portal.res_partner_grade_ready', 3)
    SILVER_GRADE = LazyRef('openerp_portal.res_partner_grade_silver', 2)
    GOLD_GRADE = LazyRef('openerp_portal.res_partner_grade_gold', 1)

    ACCOUNTING_FIRM_NOPUBLISH = (False, '0_learning')

    GRADE_ALLOW_PUBLISH = property(
        lambda self: [self.READY_GRADE, self.SILVER_GRADE, self.GOLD_GRADE]
    )
    NEXT_GRADE_COMMISSION_VALUE = property(
        lambda self: {
            self.LEARNING_GRADE: 5,
            self.READY_GRADE: 10,
            self.SILVER_GRADE: 20,
            self.GOLD_GRADE: 0
        }
    )
    DROP_ACTIVATION = 16  # FIXME: Missing in Production!

    #
    # Partnership articles (product.template)
    #
    PRODUCT_PARTNERSHIP_LEARNING = LazyRef(  # Product_template  (product.product = 6554)
        'openerp_enterprise.odoo_learning_platform_access_product_template', 6654
    )
    PRODUCT_PARTNERSHIP_OFFICIAL = LazyRef(
        'openerp_enterprise.product_partnership_yearly_product_template', 782
    )
    PRODUCT_DISCOUNT_SUBSCRIPTION_LEARNING = LazyRef(
        'openerp_enterprise.product_learning_partner_discount_product_template', 10241
    )
    PRODUCT_DISCOUNT_SUBSCRIPTION_OFFICIAL = LazyRef(
        'openerp_enterprise.product_official_partner_discount_product_template', 10240
    )
    PRODUCT_CATEGORY_PARTNERSHIP = LazyRef(
        'openerp_enterprise.product_category_partnership', 17
    )

    # Country group for success story
    EMEA = LazyRef('openerp_portal.country_group_emea', 16)
    APAC = LazyRef('openerp_portal.country_group_apac', 21)

    # Country Group For Saleteam
    US_GROUP = LazyRef('openerp_portal.country_group_north_america', 4)
    IN_GROUP = LazyRef('openerp_portal.country_group_india', 5)
    HK_GROUP = LazyRef('openerp_portal.country_group_apac', 21)

    # SalesTeams
    BE_SALES = LazyRef('openerp_portal.sale_team_be_partnership', 12)
    HK_SALES = LazyRef('openerp_portal.sale_team_hk_partnership', 60)
    US_SALES = LazyRef('openerp_portal.sale_team_us_partnership', 14)
    IN_SALES = LazyRef('openerp_portal.sale_team_in_partnership', 16)

    TEAM_BY_COUNTRY_GROUP = property(
        lambda self: {
            self.US_GROUP: self.US_SALES,
            self.IN_GROUP: self.IN_SALES,
            self.HK_GROUP: self.HK_SALES,
        }
    )

    DEFAULT_REPR_PARTNER = 1  # default partner in no saleman found by geoip

    #
    # Event types
    #
    ODOO_EXP_ID = LazyRef('openerp_website.experience_event_type', 2)  # Odoo Experience category
    ODOO_TOUR_ID = LazyRef('openerp_website.roadshow_event_type', 15)  # Odoo Tour category

    # res.partner.tag (website tags), cert ID should be order by version, the latest version last
    ODOO_CERT_IDS = property(
        lambda self: [
            self._resolve(LazyRef('openerp_portal.website_tag_odoo_10_certification', 1)),
            self._resolve(LazyRef('openerp_portal.website_tag_odoo_11_certification', 2)),
            self._resolve(LazyRef('openerp_portal.website_tag_odoo_12_certification', 3)),
            self._resolve(LazyRef('openerp_portal.website_tag_odoo_13_certification', 4)),
            self._resolve(LazyRef('openerp_portal.website_tag_odoo_14_certification', 6)),
            self._resolve(LazyRef('openerp_portal.website_tag_odoo_15_certification', 7)),
            self._resolve(LazyRef('openerp_portal.website_tag_odoo_16_certification', 8)),
            self._resolve(LazyRef('openerp_portal.website_tag_odoo_17_certification', 18)),
        ]
    )
    # Only last three version are valid certification
    ODOO_VALID_CERT_IDS = property(lambda self: self.ODOO_CERT_IDS[-3:])

    def _resolve(self, lazy_value):
        dbname = getattr(threading.current_thread(), 'dbname', None)
        # if current Thread is bound to a database and we're on a "test" host
        # without a copy of production database, then we try to dynamically
        # resolve the "xmlid" to the real "id".
        if dbname and TEST and ENV not in ('test', 'next'):
            env = _get_env()
            if env is not None:
                # we do not use env.ref() here as we we only want to get
                # the record id (leveraging the ORM cache) and we assume
                # the record still exists (env.ref() is doing an extra
                # query for that sole purpose)
                return env['ir.model.data']._xmlid_to_res_id(
                    lazy_value.xmlid, raise_if_not_found=True
                )
            else:
                # only warn during normal run if we're unable to current env
                _logger.warning(
                    'Could not find an environment to dynamically resolve a constant',
                    stack_info=True,
                )

        # for production, always use default (hardcoded) value
        return lazy_value.default

    def __getattribute__(self, name):
        """Resolve LazyRef on access."""
        value = super(Consts, self).__getattribute__(name)
        if isinstance(value, LazyRef):
            value = self._resolve(value)
        return value

    def __delitem__(self, key):
        raise NotImplementedError("'__delitem__' not supported on Consts")

    def __setitem__(self, key, val):
        raise NotImplementedError("'__setitem__' not supported on Consts")


sys.modules[__name__] = Consts()
